from csr import CSR
import numpy as np

from hypothesis import given
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph


@given(st.integers(0, 1000), st.integers(0, 1000))
def test_empty(nrows, ncols):
    csr = CSR.empty(nrows, ncols)
    assert csr.nrows == nrows
    assert csr.ncols == ncols
    assert csr.nnz == 0
    assert all(csr.rowptrs == 0)
    assert len(csr.rowptrs) == nrows + 1
    assert len(csr.colinds) == 0


@given(st.data(), st.integers(0, 1000), st.integers(0, 1000))
def test_uninitialized(data, nrows, ncols):
    sizes = data.draw(nph.arrays(np.int32, nrows, elements=st.integers(0, ncols)))
    csr = CSR.empty(nrows, ncols, sizes)
    assert csr.nrows == nrows
    assert csr.ncols == ncols
    assert csr.nnz == np.sum(sizes)
    assert len(csr.rowptrs) == nrows + 1
    assert all(csr.row_nnzs() == sizes)
    assert len(csr.colinds) == np.sum(sizes)



"""
Kernel implementing matrix operations in pure Numba.
"""


import numpy as np
from numba import njit

from csr.layout import _CSR
import csr.native_ops as _ops
from .multiply import mult_ab, mult_abt


@njit
def to_handle(csr):
    """
    Convert a native CSR to a handle.  The caller must arrange for the CSR last at
    least as long as the handle.  The handle must be explicitly released.

    Handles are opaque as far as callers are concerned.
    """
    return csr


@njit
def from_handle(h):
    """
    Convert a handle to a CSR.  The handle may be released after this is called.
    """

    return h


@njit
def release_handle(h):
    """
    Release a handle.
    """
    pass


@njit
def order_columns(h):
    """
    Sort matrix rows in increasing column order.
    """
    _ops.sort_rows(h)


@njit(nogil=True)
def mult_vec(h: _CSR, v):
    res = np.zeros(h.nrows)

    row = 0
    for i in range(h.nnz):
        # advance the row if necessary
        while i == h.rowptrs[row+1]:
            row += 1
        col = h.colinds[i]
        if h.has_values:
            res[row] += v[col] * h.values[i]
        else:
            res[row] += v[col]

    return res



import numpy as np
import pickle

from csr import CSR
from csr.test_utils import csrs, csr_slow

from pytest import mark
from hypothesis import given


@csr_slow()
@given(csrs())
def test_csr_pickle(csr):
    data = pickle.dumps(csr)
    csr2 = pickle.loads(data)

    assert csr2.nrows == csr.nrows
    assert csr2.ncols == csr.ncols
    assert csr2.nnz == csr.nnz
    assert all(csr2.rowptrs == csr.rowptrs)
    assert all(csr2.colinds == csr.colinds)
    if csr.values is not None:
        assert all(csr2.values == csr.values)
    else:
        assert csr2.values is None


@mark.skip(reason='64-bit row pointers disabled')
@csr_slow()
@given(csrs())
def test_csr64_pickle(csr):
    csr = CSR(csr.nrows, csr.ncols, csr.nnz,
              csr.rowptrs.astype(np.int64), csr.colinds, csr.values)

    data = pickle.dumps(csr)
    csr2 = pickle.loads(data)

    assert csr2.nrows == csr.nrows
    assert csr2.ncols == csr.ncols
    assert csr2.nnz == csr.nnz
    assert all(csr2.rowptrs == csr.rowptrs)
    assert csr2.rowptrs.dtype == np.int64
    assert all(csr2.colinds == csr.colinds)
    if csr.values is not None:
        assert all(csr2.values == csr.values)
    else:
        assert csr2.values is None



"""
Internal tests for Numba kernel multiplication code.
"""

import numpy as np
import scipy.sparse as sps

from csr import CSR
from csr.test_utils import csrs, csr_slow, sparse_matrices, mm_pairs
from csr.kernels.numba.multiply import _sym_mm

from pytest import mark, approx, raises
from hypothesis import given, assume, settings, HealthCheck
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph


@settings(deadline=None)
@given(mm_pairs(max_shape=(50, 1000, 50), as_csr=True))
def test_symb(pair):
    A, B = pair

    cp = np.zeros_like(A.rowptrs)

    cci = _sym_mm(A.R, B.R, cp)

    # Is everything in range?
    assert all(cci >= 0)
    assert all(cci < B.ncols)

    # Are column pointers nondecreasing?
    assert all(np.diff(cp) >= 0)

    # Do we have the right number of NNZs?
    assert len(cci) == cp[A.nrows]



import numpy as np
import scipy.sparse as sps

from csr import CSR
from csr.test_utils import csrs, csr_slow, sparse_matrices

from pytest import mark, approx, raises
from hypothesis import given, assume, settings, HealthCheck
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph


def test_csr_transpose():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)
    csc = csr.transpose()
    assert csc.nrows == csr.ncols
    assert csc.ncols == csr.nrows

    assert all(csc.rowptrs == [0, 1, 3, 4])
    assert csc.colinds.max() == 3
    assert csc.values.sum() == approx(vals.sum())

    for r, c, v in zip(rows, cols, vals):
        row = csc.row(c)
        assert row[r] == v


def test_csr_transpose_coords():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)
    csc = csr.transpose(False)
    assert csc.nrows == csr.ncols
    assert csc.ncols == csr.nrows

    assert all(csc.rowptrs == [0, 1, 3, 4])
    assert csc.colinds.max() == 3
    assert csc.values is None

    for r, c, v in zip(rows, cols, vals):
        row = csc.row(c)
        assert row[r] == 1


def test_csr_transpose_erow():
    nrows = np.random.randint(10, 1000)
    ncols = np.random.randint(10, 500)
    mat = np.random.randn(nrows, ncols)
    mat[mat <= 0] = 0
    mat[:, 0:1] = 0
    smat = sps.csr_matrix(mat)

    csr = CSR.from_scipy(smat)
    csrt = csr.transpose()
    assert csrt.nrows == ncols
    assert csrt.ncols == nrows

    s2 = csrt.to_scipy()
    smat = smat.T.tocsr()
    assert all(smat.indptr == csrt.rowptrs)

    assert np.all(s2.toarray() == smat.toarray())


@given(sparse_matrices())
def test_csr_transpose_many(smat):
    nrows, ncols = smat.shape
    csr = CSR.from_scipy(smat)
    csrt = csr.transpose()
    assert csrt.nrows == ncols
    assert csrt.ncols == nrows
    assert csrt.nnz == csr.nnz

    s2 = csrt.to_scipy()
    smat = smat.T.tocsr()
    assert all(smat.indptr == csrt.rowptrs)

    assert np.all(s2.toarray() == smat.toarray())



"""
CSR test utilities.
"""

import numpy as np
import scipy.sparse as sps

from hypothesis import settings, HealthCheck
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph

from .csr import CSR


def fractions(**kwargs):
    return st.floats(0, 1, **kwargs)


@st.composite
def csrs(draw, nrows=None, ncols=None, nnz=None, values=None):
    "Draw CSR matrices by generating COO data."
    if ncols is None:
        ncols = draw(st.integers(1, 100))
    elif not isinstance(ncols, int):
        ncols = draw(ncols)

    if nrows is None:
        nrows = draw(st.integers(1, 100))
    elif not isinstance(nrows, int):
        nrows = draw(nrows)

    if nnz is None:
        nnz = draw(st.integers(0, int(np.ceil(nrows * ncols * 0.5))))
    elif not isinstance(nnz, int):
        nnz = draw(nnz)

    coords = draw(nph.arrays(np.int32, nnz, elements=st.integers(0, nrows*ncols - 1), unique=True))
    rows = np.mod(coords, nrows, dtype=np.int32)
    cols = np.floor_divide(coords, nrows, dtype=np.int32)
    if values is None:
        values = draw(st.booleans())
    if values:
        vals = draw(nph.arrays(np.float64, nnz, elements=st.floats(-10, 10)))
    else:
        vals = None
    return CSR.from_coo(rows, cols, vals, (nrows, ncols))


@st.composite
def sparse_matrices(draw, max_shape=(1000, 1000), density=fractions(), format='csr'):
    ubr, ubc = max_shape
    rows = draw(st.integers(1, ubr))
    cols = draw(st.integers(1, ubc))
    dens = draw(density)
    return sps.random(rows, cols, dens, format=format)


@st.composite
def mm_pairs(draw, max_shape=(100, 100, 100), as_csr=False):
    "Draw multipliable pairs of matrices"
    mr, mm, mc = max_shape
    rows = draw(st.integers(1, mr))
    mids = draw(st.integers(1, mm))
    cols = draw(st.integers(1, mc))
    dA = draw(st.floats(0.001, 0.9))
    dB = draw(st.floats(0.001, 0.9))

    A = sps.random(rows, mids, dA, format='csr')
    B = sps.random(mids, cols, dB, format='csr')

    if as_csr:
        return CSR.from_scipy(A), CSR.from_scipy(B)
    else:
        return A, B


def matrices(max_shape=(100, 100), dtype='f8'):
    "Draw dense matrices"
    ubr, ubc = max_shape
    return nph.arrays(dtype, st.tuples(st.integers(1, ubr), st.integers(1, ubc)))


def csr_slow(divider=2):
    dft = settings.default
    return settings(dft, deadline=None, suppress_health_check=HealthCheck.all(),
                    max_examples=dft.max_examples // divider)



import os
import warnings
from contextlib import contextmanager
from importlib import import_module
import threading

kernels = {}
__all__ = [
    'releasing',
    'set_kernel',
    'use_kernel',
    'get_kernel',
]


class ActiveKernel(threading.local):
    def __init__(self):
        self.__dict__.update({'active_name': None})

    @property
    def active(self):
        kern = getattr(self, '_active', None)
        if kern is None:
            return _default_kernel()
        else:
            return kern

    def set_active(self, kern):
        self._active = kern


__cached_default = None
__active = ActiveKernel()


@contextmanager
def releasing(h, k):
    try:
        yield h
    finally:
        k.release_handle(h)


def set_kernel(name):
    """
    Set the default kernel.  It is very rare to need to use this — letting CSR select
    its default kernel, or configuring the kernel through the ``CSR_KERNEL`` environment
    variable, is the best option for the vast majority of applications.  This is here
    primarily to enable test code to switch kernels.

    This function does **not** change the kernel exposed by importing ``csr.kernel``,
    which is typically used by compiled Numba functions.  It only changes the kernel
    returned by :py:func:`get_kernel` and used by the pure-Python APIs,

    Args:
        name(str):
            The name of the kernel.
    """

    if name is None:
        __active.set_active(None)
    else:
        __active.set_active(get_kernel(name))


@contextmanager
def use_kernel(name):
    """
    Context manager to run code with a specified (thread-local) kernel.  It calls
    :py:func:`set_kernel`, and restores the previously-active kernel when the context
    exits.
    """
    old = __active.active_name
    try:
        set_kernel(name)
        yield
    finally:
        set_kernel(old)


def get_kernel(name=None):
    """
    Get a kernel.

    Args:
        name(str or None):
            The name of the kernel.  If ``None``, returns the current default kernel.
    """
    if name is None:
        return __active.active

    kern = kernels.get(name, None)
    if not kern:
        mod_name = f'{__name__}.{name}'
        kern = import_module(mod_name)
        kernels[name] = kern
    return kern


def _initialize(name=None):
    global __cached_default
    if __cached_default:
        warnings.warn('default kernel already initialized')

    if name:
        k = import_module(f'csr.kernels.{name}')
    elif 'CSR_KERNEL' in os.environ:
        _env_kernel = os.environ['CSR_KERNEL']
        k = import_module(f'csr.kernels.{_env_kernel}')
    else:
        try:
            import csr.kernels.mkl as k
        except ImportError:
            import csr.kernels.numba as k

    __cached_default = k


def _default_kernel():
    if not __cached_default:
        _initialize()

    return __cached_default



import numpy as np
from numba import njit
import numba.types as nt
from numba.experimental import jitclass

from csr import _CSR
from ._api import *
from csr.native_ops import make_empty

__all__ = [
    'mkl_h',
    'to_handle',
    'from_handle',
    'release_handle'
]


@jitclass([
    ('H', nt.intp),
    ('nrows', nt.intc),
    ('ncols', nt.intc),
    ('values', nt.float64[::1])
])
class mkl_h:
    """
    Type for MKL handles.  Do not use this directly.
    """
    def __init__(self, H, nrows, ncols, vs):
        self.H = H
        self.nrows = nrows
        self.ncols = ncols
        self.values = vs


@njit
def to_handle(csr: _CSR) -> mkl_h:
    if csr.nnz == 0:
        # empty matrices don't really work
        return mkl_h(0, csr.nrows, csr.ncols, np.zeros(0))

    _sp = ffi.from_buffer(csr.rowptrs)
    _cols = ffi.from_buffer(csr.colinds)
    if csr.has_values:
        vs = csr.values
    else:
        vs = np.ones(csr.nnz)
    _vals = ffi.from_buffer(vs)
    h = lk_mkl_spcreate(csr.nrows, csr.ncols, _sp, _cols, _vals)
    lk_mkl_spopt(h)
    return mkl_h(h, csr.nrows, csr.ncols, vs)


@njit
def from_handle(h: mkl_h) -> _CSR:
    if not h.H:
        return make_empty(h.nrows, h.ncols)

    rvp = lk_mkl_spexport_p(h.H)
    if rvp is None:
        return None

    nrows = lk_mkl_spe_nrows(rvp)
    ncols = lk_mkl_spe_ncols(rvp)

    sp = lk_mkl_spe_row_sp(rvp)
    ep = lk_mkl_spe_row_ep(rvp)
    cis = lk_mkl_spe_colinds(rvp)
    vs = lk_mkl_spe_values(rvp)

    rowptrs = np.zeros(nrows + 1, dtype=np.intc)
    nnz = 0
    for i in range(nrows):
        nnz += ep[i] - sp[i]
        rowptrs[i+1] = nnz

    colinds = np.zeros(nnz, dtype=np.intc)
    values = np.zeros(nnz)

    for i in range(nrows):
        rs = rowptrs[i]
        re = rowptrs[i+1]
        ss = sp[i]
        for j in range(re - rs):
            colinds[rs + j] = cis[ss + j]
            values[rs + j] = vs[ss + j]

    lk_mkl_spe_free(rvp)

    return _CSR(nrows, ncols, nnz, rowptrs, colinds, values)


@njit
def order_columns(h):
    """
    Sort matrix rows in increasing column order.
    """
    if h.H:
        lk_mkl_sporder(h.H)


@njit
def release_handle(h: mkl_h):
    if h.H:
        lk_mkl_spfree(h.H)
    h.H = 0



import numpy as np
import scipy.sparse as sps

from csr import CSR
from csr.test_utils import csrs, csr_slow, sparse_matrices

from pytest import mark, approx, raises
from hypothesis import given, assume, settings, HealthCheck
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph


@mark.parametrize('copy', [True, False])
def test_csr_from_sps(copy):
    "Test creating a CSR from a SciPy matrix"
    # initialize sparse matrix
    mat = np.random.randn(10, 5)
    mat[mat <= 0] = 0
    smat = sps.csr_matrix(mat)
    # make sure it's sparse
    assert smat.nnz == np.sum(mat > 0)

    csr = CSR.from_scipy(smat, copy=copy)
    assert csr.nnz == smat.nnz
    assert csr.nrows == smat.shape[0]
    assert csr.ncols == smat.shape[1]

    assert all(csr.rowptrs == smat.indptr)
    assert all(csr.colinds == smat.indices)
    assert all(csr.values == smat.data)
    assert isinstance(csr.rowptrs, np.ndarray)
    assert isinstance(csr.colinds, np.ndarray)
    assert isinstance(csr.values, np.ndarray)


@given(sparse_matrices(format='csr'), st.booleans())
def test_csr_from_sps_csr(smat, copy):
    "Test creating a CSR from a SciPy CSR matrix"
    csr = CSR.from_scipy(smat, copy=copy)
    assert csr.nnz == smat.nnz
    assert csr.nrows == smat.shape[0]
    assert csr.ncols == smat.shape[1]

    assert all(csr.rowptrs == smat.indptr)
    assert all(csr.colinds == smat.indices)
    assert all(csr.values == smat.data)
    assert isinstance(csr.rowptrs, np.ndarray)
    assert isinstance(csr.colinds, np.ndarray)
    if csr.nnz > 0:
        assert isinstance(csr.values, np.ndarray)


def test_csr_is_numpy_compatible():
    # initialize sparse matrix
    mat = np.random.randn(10, 5)
    mat[mat <= 0] = 0
    smat = sps.csr_matrix(mat)
    # make sure it's sparse
    assert smat.nnz == np.sum(mat > 0)

    csr = CSR.from_scipy(smat)

    d2 = csr.values * 10
    assert d2 == approx(smat.data * 10)


def test_csr_from_coo():
    "Make a CSR from COO data"
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)
    assert csr.nrows == 4
    assert csr.ncols == 3
    assert csr.nnz == 4
    assert csr.values == approx(vals)


def test_csr_from_coo_rand():
    for i in range(100):
        coords = np.random.choice(np.arange(50 * 100, dtype=np.int32), 1000, False)
        rows = np.mod(coords, 100, dtype=np.int32)
        cols = np.floor_divide(coords, 100, dtype=np.int32)
        vals = np.random.randn(1000)

        csr = CSR.from_coo(rows, cols, vals, (100, 50))
        rowinds = csr.rowinds()
        assert csr.nrows == 100
        assert csr.ncols == 50
        assert csr.nnz == 1000

        for i in range(100):
            sp = csr.rowptrs[i]
            ep = csr.rowptrs[i+1]
            assert ep - sp == np.sum(rows == i)
            points, = np.nonzero(rows == i)
            assert len(points) == ep - sp
            po = np.argsort(cols[points])
            points = points[po]
            assert all(np.sort(csr.colinds[sp:ep]) == cols[points])
            assert all(np.sort(csr.row_cs(i)) == cols[points])
            assert all(csr.values[np.argsort(csr.colinds[sp:ep]) + sp] == vals[points])
            assert all(rowinds[sp:ep] == i)

            row = np.zeros(50)
            row[cols[points]] = vals[points]
            assert np.sum(csr.row(i)) == approx(np.sum(vals[points]))
            assert all(csr.row(i) == row)


def test_csr_from_coo_novals():
    for i in range(50):
        coords = np.random.choice(np.arange(50 * 100, dtype=np.int32), 1000, False)
        rows = np.mod(coords, 100, dtype=np.int32)
        cols = np.floor_divide(coords, 100, dtype=np.int32)

        csr = CSR.from_coo(rows, cols, None, (100, 50))
        assert csr.nrows == 100
        assert csr.ncols == 50
        assert csr.nnz == 1000

        for i in range(100):
            sp = csr.rowptrs[i]
            ep = csr.rowptrs[i+1]
            assert ep - sp == np.sum(rows == i)
            points, = np.nonzero(rows == i)
            po = np.argsort(cols[points])
            points = points[po]
            assert all(np.sort(csr.colinds[sp:ep]) == cols[points])
            assert np.sum(csr.row(i)) == len(points)


def test_csr_to_sps():
    # initialize sparse matrix
    mat = np.random.randn(10, 5)
    mat[mat <= 0] = 0
    # get COO
    smat = sps.coo_matrix(mat)
    # make sure it's sparse
    assert smat.nnz == np.sum(mat > 0)

    csr = CSR.from_coo(smat.row, smat.col, smat.data, shape=smat.shape)
    assert csr.nnz == smat.nnz
    assert csr.nrows == smat.shape[0]
    assert csr.ncols == smat.shape[1]

    smat2 = csr.to_scipy()
    assert sps.isspmatrix(smat2)
    assert sps.isspmatrix_csr(smat2)

    for i in range(csr.nrows):
        assert smat2.indptr[i] == csr.rowptrs[i]
        assert smat2.indptr[i+1] == csr.rowptrs[i+1]
        sp = smat2.indptr[i]
        ep = smat2.indptr[i+1]
        assert all(smat2.indices[sp:ep] == csr.colinds[sp:ep])
        assert all(smat2.data[sp:ep] == csr.values[sp:ep])



import logging
import numpy as np

from csr import CSR
from csr.test_utils import sparse_matrices

from pytest import approx
from hypothesis import given, settings
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph

_log = logging.getLogger(__name__)


@settings(deadline=None)
@given(st.data())
def test_mult_vec(kernel, data):
    mat = data.draw(sparse_matrices((100, 100)))
    md = mat.toarray()
    csra = CSR.from_scipy(mat)
    # TODO make the test work with larger values
    vals = st.floats(-100, 100)
    v = data.draw(nph.arrays(np.float64, csra.ncols, elements=vals))

    prod = csra.mult_vec(v)
    assert prod.shape == (csra.nrows,)

    v2 = md @ v

    assert prod == approx(v2, nan_ok=True, rel=1.0e-5)


@settings(deadline=None)
@given(st.data())
def test_mult_vec_novalue(kernel, data):
    mat = data.draw(sparse_matrices())
    csra = CSR.from_scipy(mat, True)

    csra.drop_values()
    mat.data[:] = 1.0

    vals = st.floats(-100, 100)
    v = data.draw(nph.arrays(np.float64, csra.ncols, elements=vals))

    prod = csra.mult_vec(v)
    assert prod.shape == (csra.nrows,)

    v2 = mat @ v

    assert prod == approx(v2, nan_ok=True)



"""
Default kernel interface for Numba.
"""

from csr.kernels import _default_kernel

_kernel = _default_kernel()

name = _kernel.__name__
to_handle = _kernel.to_handle
from_handle = _kernel.from_handle
release_handle = _kernel.release_handle
order_columns = _kernel.order_columns
mult_ab = _kernel.mult_ab
mult_abt = _kernel.mult_abt
mult_vec = _kernel.mult_vec



import logging

from hypothesis import settings
from pytest import fixture
from csr.kernels import use_kernel, get_kernel

# turn off Numba logging
logging.getLogger('numba').setLevel(logging.INFO)

KERNELS = ["scipy", "numba"]
try:
    import csr.kernels.mkl  # noqa: F401
    KERNELS.append("mkl")
except ImportError:
    pass  # no MKL available


# set up fixtures
@fixture(scope="module", params=KERNELS)
def kernel(request):
    """
    Fixture for variable CSR kernels.  This fixture is parameterized, so if you
    write a test function with a parameter ``kernel`` as its first parameter, it
    will be called once for each kernel under active test.
    """
    with use_kernel(request.param):
        yield get_kernel()


# set up profiles
settings.register_profile('default', deadline=756)
settings.register_profile('large', max_examples=5000)
settings.register_profile('nojit', max_examples=10, deadline=None)
settings.load_profile('default')



"""
Matrix multiplication using the SMMP algorithm [SMMP]_.

.. [SMMP] Bank, R. E., & Douglas, C. C. (1993). Sparse matrix multiplication package (SMMP).
    _Advances in Computational Mathematics_, 1(1), 127–137. <http://dx.doi.org/10.1007/BF02070824>
"""

import numpy as np
from numba import njit
from csr.layout import _CSR
from csr.native_ops import row_extent, transpose


@njit(nogil=True)
def mult_ab(a_h, b_h):
    """
    Multiply matrices A and B.

    Args:
        a_h: the handle of matrix A
        b_h: the handle of matrix B

    Returns:
        the handle of the product; it must be released when no longer needed.
    """

    assert a_h.ncols == b_h.nrows

    c_rp = np.zeros(a_h.nrows + 1, np.intc)

    # step 1: symbolic multiplication
    c_ci = _sym_mm(a_h, b_h, c_rp)
    c_nnz = c_rp[a_h.nrows]

    # step 2: numeric multiplication
    c_vs = _num_mm(a_h, b_h, c_rp, c_ci)

    # build the result
    return _CSR(a_h.nrows, b_h.ncols, c_nnz, c_rp, c_ci, c_vs)


@njit(nogil=True)
def mult_abt(a_h, b_h):
    """
    Multiply matrices A and B^T.

    Args:
        a_h: the handle of matrix A
        b_h: the handle of matrix B

    Returns:
        the handle of the product; it must be released when no longer needed.
    """
    assert a_h.ncols == b_h.ncols

    # transpose B
    bt_h = transpose(b_h, True)
    return mult_ab(a_h, bt_h)


@njit
def _sym_mm(a_h, b_h, c_rp):
    wlen = max(a_h.nrows, a_h.ncols, b_h.ncols)
    index = np.full(wlen, -1, np.intc)
    c_len = max(a_h.nnz, b_h.nnz)
    c_ci = np.zeros(c_len, np.intc)
    c_pos = 0

    for i in range(a_h.nrows):
        istart = wlen
        length = 0

        # Pass 1: count and link columns in row
        a_rs, a_re = row_extent(a_h, i)
        for jj in range(a_rs, a_re):
            j = a_h.colinds[jj]
            b_rs, b_re = row_extent(b_h, j)
            for kk in range(b_rs, b_re):
                k = b_h.colinds[kk]
                if index[k] < 0:
                    index[k] = istart
                    istart = k
                    length += 1

        # Make sure we have enough length
        while c_pos + length > c_len:
            ocl = c_len
            c_len = c_len + c_len // 2
            ci2 = np.empty(c_len, np.intc)
            ci2[:ocl] = c_ci
            c_ci = ci2

        # Pass 2: fill in the rows
        c_rp[i+1] = c_rp[i] + length
        for j in range(c_rp[i], c_rp[i+1]):
            c_ci[j] = istart
            istart = index[istart]
            index[c_ci[j]] = -1
        c_pos += length

    return c_ci[:c_pos]


@njit
def _num_mm(a_h, b_h, c_rp, c_ci):
    # set up work array
    wlen = max(a_h.nrows, a_h.ncols, b_h.ncols)
    work = np.zeros(wlen)
    # initalize result array
    c_vs = np.zeros(len(c_ci))

    for i in range(a_h.nrows):
        a_rs, a_re = row_extent(a_h, i)
        for jj in range(a_rs, a_re):
            j = a_h.colinds[jj]
            av = a_h.values[jj]

            b_rs, b_re = row_extent(b_h, j)
            for kk in range(b_rs, b_re):
                k = b_h.colinds[kk]
                work[k] += av * b_h.values[kk]

        c_rs = c_rp[i]
        c_re = c_rp[i+1]
        for jj in range(c_rs, c_re):
            j = c_ci[jj]
            c_vs[jj] = work[j]
            work[j] = 0

    return c_vs



from csr.kernels import _default_kernel, get_kernel, use_kernel
default_kernel = _default_kernel()


def test_get_default():
    k = get_kernel()
    assert k is not None
    assert k is default_kernel

    # make sure the default is right
    try:
        import csr.kernels.mkl  # noqa: F401
        assert k.__name__ == 'csr.kernels.mkl'
    except ImportError:
        assert k.__name__ == 'csr.kernels.numba'


def test_get_scipy():
    k = get_kernel('scipy')
    assert k.__name__ == 'csr.kernels.scipy'


def test_get_numba():
    k = get_kernel('numba')
    assert k.__name__ == 'csr.kernels.numba'


def test_with_scipy():
    with use_kernel('scipy'):
        k = get_kernel()
        assert k.__name__ == 'csr.kernels.scipy'

    k = get_kernel()
    assert k is default_kernel



import numpy as np
import scipy.sparse as sps

from csr import CSR
from csr.test_utils import csrs, csr_slow, sparse_matrices

from pytest import mark, approx, raises
from hypothesis import given, assume, settings, HealthCheck
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph


@csr_slow()
@given(st.data())
def test_subset_rows(data):
    nrows = data.draw(st.integers(5, 100))
    ncols = data.draw(st.integers(1, 100))
    dens = data.draw(st.floats(0, 1))
    beg = data.draw(st.integers(0, nrows - 1))
    end = data.draw(st.integers(beg, nrows - 1))

    spm = sps.random(nrows, ncols, dens, format='csr')
    csr = CSR.from_scipy(spm)

    m2 = csr.subset_rows(beg, end)
    assert m2.nrows == end - beg

    for i in range(m2.nrows):
        assert all(m2.row_cs(i) == csr.row_cs(beg+i))
        assert all(m2.row_vs(i) == csr.row_vs(beg+i))


@given(csrs())
def test_sort_rows(csr):
    tv = np.ones(csr.ncols)
    x1 = csr.mult_vec(tv)
    csr.sort_rows()
    assert all(all(np.diff(csr.row_cs(i)) > 0) for i in range(csr.nrows))
    x2 = csr.mult_vec(tv)
    assert x2 == approx(x1)


@given(csrs())
def test_kernel_sort_rows(kernel, csr):
    tv = np.ones(csr.ncols)
    x1 = csr.mult_vec(tv)
    h = kernel.to_handle(csr.R)
    kernel.order_columns(h)
    c2 = kernel.from_handle(h)
    c2 = CSR(R=c2)
    kernel.release_handle(h)
    assert all(all(np.diff(c2.row_cs(i)) > 0) for i in range(csr.nrows))
    x2 = c2.mult_vec(tv)
    assert x2 == approx(x1)


@csr_slow()
@given(sparse_matrices())
def test_mean_center(spm):
    # assume(spm.nnz >= 10)
    csr = CSR.from_scipy(spm)

    m2 = csr.normalize_rows('center')
    assert len(m2) == csr.nrows
    rnnz = csr.row_nnzs()

    for i in range(csr.nrows):
        vs = csr.row_vs(i)
        spr = spm[i, :].toarray()
        if rnnz[i] > 0:
            assert m2[i] == approx(np.sum(spr) / rnnz[i])
            assert np.mean(vs) == approx(0.0)
            assert vs + m2[i] == approx(spr[0, csr.row_cs(i)])


@csr_slow()
@given(sparse_matrices())
def test_unit_norm(spm):
    # assume(spm.nnz >= 10)
    csr = CSR.from_scipy(spm)

    m2 = csr.normalize_rows('unit')
    assert len(m2) == csr.nrows

    for i in range(csr.nrows):
        vs = csr.row_vs(i)
        if len(vs) > 0:
            assert np.linalg.norm(vs) == approx(1.0)
            assert vs * m2[i] == approx(spm.getrow(i).toarray()[0, csr.row_cs(i)])


@csr_slow()
@given(sparse_matrices())
def test_filter(mat):
    assume(mat.nnz > 0)
    assume(not np.all(mat.data <= 0))  # we have to have at least one to retain
    csr = CSR.from_scipy(mat)
    csrf = csr.filter_nnzs(csr.values > 0)
    assert all(csrf.values > 0)
    assert csrf.nnz <= csr.nnz

    for i in range(csr.nrows):
        spo, epo = csr.row_extent(i)
        spf, epf = csrf.row_extent(i)
        assert epf - spf <= epo - spo

    d1 = csr.to_scipy().toarray()
    df = csrf.to_scipy().toarray()
    d1[d1 < 0] = 0
    assert df == approx(d1)



"""
SciPy "kernel".  This kernel is not Numba-compatible, and will never be
selected as the default.  It primarily exists for ease in testing and
benchmarking CSR operations.
"""

import numpy as np
from scipy.sparse import csr_matrix
from csr import CSR, _CSR


def to_handle(csr: _CSR):
    values = csr.values if csr.has_values else np.ones(csr.nnz)
    return csr_matrix((values, csr.colinds, csr.rowptrs), (csr.nrows, csr.ncols))


def from_handle(h):
    m: csr_matrix = h.tocsr()
    nr, nc = m.shape
    return _CSR(nr, nc, m.nnz, m.indptr, m.indices, m.data)


def order_columns(h):
    h.sort_indices()


def release_handle(h):
    pass


def mult_ab(A, B):
    return A @ B


def mult_abt(A, B):
    return A @ B.T


def mult_vec(A, v):
    return A @ v



"""
Python API for CSR matrices.
"""

import numpy as np
import scipy.sparse as sps

from csr.kernels import get_kernel, releasing
from csr import native_ops as _ops
from csr.layout import EMPTY_VALUES, _CSR


def _csr_delegate(name):
    def func(self):
        return getattr(self.R, name)

    return property(func)


class CSR:
    """
    Simple compressed sparse row matrix.  This is like :py:class:`scipy.sparse.csr_matrix`, with
    a couple of useful differences:

    * The value array is optional, for cases in which only the matrix structure is required.
    * The value array, if present, is always double-precision.

    You generally don't want to create this class yourself with the constructor.  Instead, use one
    of its class methods.

    It is backed by separate storage type (py:class:`csr._CSR`) that can be passed around through
    Numba-compiled functions, and nopython compiled equivalents of many of its methods
    are available as functions in the :py:mod:`csr.native_ops` module that take the
    underlying tuple (accessible by :py:attr:`R`) as a parameter.

    If you need to pass an instance off to a Numba-compiled function, use :py:attr:`R`::

        _some_numba_fun(csr.R)

    Attributes:
        R(_CSR): the named tuple containing the actual matrix data
        nrows(int): the number of rows.
        ncols(int): the number of columns.
        nnz(int): the number of entries.
        rowptrs(numpy.ndarray): the row pointers.
        colinds(numpy.ndarray): the column indices.
        values(numpy.ndarray or None): the values
    """
    __slots__ = ['R']

    def __init__(self, nrows=None, ncols=None, nnz=None, ptrs=None, inds=None, vals=None, R=None):
        if R is not None:
            assert R.colinds.dtype == np.int32
            self.R = R
        else:
            assert inds.dtype == np.int32
            self.R = _CSR(nrows, ncols, nnz, ptrs, inds, vals)

    @classmethod
    def empty(cls, nrows, ncols, row_nnzs=None):
        """
        Create an uninitialized CSR matrix.

        Args:
            nrows(int): the number of rows.
            ncols(int): the number of columns.
            row_nnzs(array-like):
                the number of nonzero entries for each row, or None for an empty matrix.
        """
        if row_nnzs is not None:
            assert len(row_nnzs) == nrows
            return cls(R=_ops.make_unintialized(nrows, ncols, row_nnzs))
        else:
            return cls(R=_ops.make_empty(nrows, ncols))

    @classmethod
    def from_coo(cls, rows, cols, vals, shape=None, rpdtype=np.intc):
        """
        Create a CSR matrix from data in COO format.

        Args:
            rows(array-like): the row indices.
            cols(array-like): the column indices.
            vals(array-like): the data values; can be ``None``.
            shape(tuple): the array shape, or ``None`` to infer from row & column indices.
        """
        if shape is not None:
            nrows, ncols = shape
            assert np.max(rows, initial=0) < nrows
            assert np.max(cols, initial=0) < ncols
        else:
            nrows = np.max(rows) + 1
            ncols = np.max(cols) + 1

        nnz = len(rows)
        assert len(cols) == nnz
        assert vals is None or len(vals) == nnz

        rowptrs = np.zeros(nrows + 1, dtype=rpdtype)
        align = np.full(nnz, -1, dtype=rpdtype)

        _ops._csr_align(rows, nrows, rowptrs, align)

        cols = cols[align].copy()
        vals = vals[align].copy() if vals is not None else None

        return cls(nrows, ncols, nnz, rowptrs, cols, vals)

    @classmethod
    def from_scipy(cls, mat, copy=True):
        """
        Convert a scipy sparse matrix to a CSR.

        Args:
            mat(scipy.sparse.spmatrix): a SciPy sparse matrix.
            copy(bool): if ``False``, reuse the SciPy storage if possible.

        Returns:
            CSR: a CSR matrix.
        """
        if not sps.isspmatrix_csr(mat):
            mat = mat.tocsr(copy=copy)
        rp = np.require(mat.indptr, np.intc, 'C')
        if copy and rp is mat.indptr:
            rp = rp.copy()
        cs = np.require(mat.indices, np.intc, 'C')
        if copy and cs is mat.indices:
            cs = cs.copy()
        vs = mat.data.copy() if copy else mat.data
        return cls(mat.shape[0], mat.shape[1], mat.nnz, rp, cs, vs)

    def to_scipy(self):
        """
        Convert a CSR matrix to a SciPy :py:class:`scipy.sparse.csr_matrix`.  Avoids copying
        if possible.

        Args:
            self(CSR): A CSR matrix.

        Returns:
            scipy.sparse.csr_matrix:
                A SciPy sparse matrix with the same data.
        """
        values = self.values
        if values is None:
            values = np.full(self.nnz, 1.0)
        return sps.csr_matrix((values, self.colinds, self.rowptrs), shape=(self.nrows, self.ncols))

    nrows = _csr_delegate('nrows')
    ncols = _csr_delegate('ncols')
    nnz = _csr_delegate('nnz')
    rowptrs = _csr_delegate('rowptrs')
    colinds = _csr_delegate('colinds')

    @property
    def values(self):
        return self.R.values if self.R.has_values else None

    @values.setter
    def values(self, vs: np.ndarray):
        if vs is None:
            self.R.values = EMPTY_VALUES
            self.R.has_values = False
        else:
            if len(vs) < self.nnz:
                raise ValueError('value array too small')
            elif len(vs) > self.nnz:
                vs = vs[:self.nnz]
            self.R.values = vs
            self.R.has_values = True

    def copy(self, include_values=True, *, copy_structure=True):
        """
        Create a copy of this CSR.

        Args:
            include_values(bool): whether to copy the values or only the structure.
            copy_structure(bool):
                whether to copy the structure (index & pointers) or share with the original matrix.
        """
        values = self.values
        if include_values and values is not None:
            values = np.copy(values)
        else:
            values = None
        rps = self.rowptrs
        cis = self.colinds
        if copy_structure:
            rps = np.copy(rps)
            cis = np.copy(cis)
        return CSR(self.nrows, self.ncols, self.nnz,
                   rps, cis, values)

    def sort_rows(self):
        """
        Sort the rows of this matrix in column order.  This is an **in-place operation**.
        """
        _ops.sort_rows(self.R)

    def subset_rows(self, begin, end):
        """
        Subset the rows in this matrix.
        """
        return CSR(R=_ops.subset_rows(self.R, begin, end))

    def rowinds(self) -> np.ndarray:
        """
        Get the row indices from this array.  Combined with :py:attr:`colinds` and
        :py:attr:`values`, this can form a COO-format sparse matrix.
        """
        # we have to upgrade to Numba for this one - too slow in Python
        return _ops.rowinds(self.R)

    def row(self, row):
        """
        Return a row of this matrix as a dense ndarray.

        Args:
            row(int): the row index.

        Returns:
            numpy.ndarray:
                the row, with 0s in the place of missing values.  If the CSR only
                stores matrix structure, the returned vector has 1s where the CSR
                records an entry.
        """
        return _ops.row(self.R, row)

    def row_extent(self, row):
        """
        Get the extent of a row in the underlying column index and value arrays.

        Args:
            row(int): the row index.

        Returns:
            tuple: ``(s, e)``, where the row occupies positions :math:`[s, e)` in the
            CSR data.
        """
        return _ops.row_extent(self.R, row)

    def row_cs(self, row):
        """
        Get the column indcies for the stored values of a row.
        """
        return _ops.row_cs(self.R, row)

    def row_vs(self, row):
        """
        Get the stored values of a row.  If only the matrix structure is stored, this
        returns a vector of 1s.
        """
        return _ops.row_vs(self.R, row)

    def row_nnzs(self):
        """
        Get a vector of the number of nonzero entries in each row.

        Returns:
            numpy.ndarray: the number of nonzero entries in each row.
        """
        return np.diff(self.rowptrs)

    def normalize_rows(self, normalization):
        """
        Normalize the rows of the matrix.

        .. note:: The normalization *ignores* missing values instead of treating
                  them as 0.

        Args:
            normalization(str):
                The normalization to perform. Can be one of:

                * ``'center'`` - center rows about the mean
                * ``'unit'`` - convert rows to a unit vector

        Returns:
            numpy.ndarray:
                The normalization values for each row.
        """
        if normalization == 'center':
            return _ops.center_rows(self.R)
        elif normalization == 'unit':
            return _ops.unit_rows(self.R)
        else:
            raise ValueError('unknown normalization: ' + normalization)

    def transpose(self, values=True):
        """
        Transpose a CSR matrix.

        Args:
            values(bool): whether to include the values in the transpose.

        Returns:
            CSR: the transpose of this matrix (or, equivalently, this matrix in CSC format).
        """

        r2 = _ops.transpose(self.R, values)
        return CSR(R=r2)

    def filter_nnzs(self, filt):
        """
        Filter the values along the full NNZ axis.

        Args:
            filt(ndarray):
                a logical array of length :attr:`nnz` that indicates the values to keep.

        Returns:
            CSR: The filtered sparse matrix.
        """
        if len(filt) != self.nnz:
            raise ValueError('filter has length %d, expected %d' % (len(filt), self.nnz))
        rps2 = np.zeros_like(self.rowptrs)
        for i in range(self.nrows):
            sp, ep = self.row_extent(i)
            rlen = np.sum(filt[sp:ep])
            rps2[i+1] = rps2[i] + rlen

        nnz2 = rps2[-1]
        assert nnz2 == np.sum(filt)

        cis2 = self.colinds[filt]
        vs = self.values
        vs2 = None if vs is None else vs[filt]

        return CSR(self.nrows, self.ncols, nnz2, rps2, cis2, vs2)

    def multiply(self, other, *, transpose=False):
        """
        Multiply this matrix by another.

        Args:
            other(CSR): the other matrix.
            transpose(bool): if ``True``, compute :math:`AB^{T}` instead of :math:`AB`.

        Returns
            CSR: the product of the two matrices.
        """
        if transpose:
            assert self.ncols == other.ncols
        else:
            assert self.ncols == other.nrows

        K = get_kernel()
        with releasing(K.to_handle(self.R), K) as a_h:
            with releasing(K.to_handle(other.R), K) as b_h:
                if transpose:
                    c_h = K.mult_abt(a_h, b_h)
                else:
                    c_h = K.mult_ab(a_h, b_h)
                with releasing(c_h, K):
                    crepr = K.from_handle(c_h)

        return CSR(R=crepr)

    def mult_vec(self, v):
        """
        Multiply this matrix by a vector.

        Args:
            other(numpy.ndarray): A vector, of length `ncols`.

        Returns:
            numpy.ndarray: :math:`A\\vec{x}`, as a vector.
        """
        assert v.shape == (self.ncols,)
        K = get_kernel()
        with releasing(K.to_handle(self.R), K) as h:
            return K.mult_vec(h, v)

    def drop_values(self):
        """
        Remove the value array from this CSR.  This is an **in-place** operation.
        """
        self.R.values = EMPTY_VALUES
        self.R.has_values = False

    def fill_values(self, value):
        """
        Fill the values of this CSR with the specified value.  If the CSR is
        structure-only, a value array is added.  This is an **in-place** operation.
        """
        self.R.values = np.full(self.nnz, value, dtype='float64')
        self.R.has_values = True

    def __str__(self):
        return '<CSR {}x{} ({} nnz)>'.format(self.nrows, self.ncols, self.nnz)

    def __repr__(self):
        repr = '<CSR {}x{} ({} nnz)'.format(self.nrows, self.ncols, self.nnz)
        repr += ' {\n'
        repr += '  rowptrs={}\n'.format(self.rowptrs)
        repr += '  colinds={}\n'.format(self.colinds)
        repr += '  values={}\n'.format(self.values)
        repr += '}'
        return repr

    def __getstate__(self):
        return dict(shape=(self.nrows, self.ncols), nnz=self.nnz,
                    rowptrs=self.rowptrs, colinds=self.colinds, values=self.values)

    def __setstate__(self, state):
        nrows, ncols = state['shape']
        nnz = state['nnz']
        rps = state['rowptrs']
        cis = state['colinds']
        vs = state['values']
        self.R = _CSR(nrows, ncols, nnz, rps, cis, vs)



import numpy as np
from numba import njit

from csr.layout import EMPTY_VALUES
from ._api import *
from .handle import mkl_h

__all__ = [
    'mult_ab',
    'mult_abt'
]


@njit(nogil=True)
def mult_ab(a_h, b_h):
    if a_h.H and b_h.H:
        h = lk_mkl_spmab(a_h.H, b_h.H)
    else:
        h = 0
    return mkl_h(h, a_h.nrows, b_h.ncols, EMPTY_VALUES)


@njit(nogil=True)
def mult_abt(a_h, b_h):
    if a_h.H and b_h.H:
        h = lk_mkl_spmabt(a_h.H, b_h.H)
    else:
        h = 0
    return mkl_h(h, a_h.nrows, b_h.nrows, EMPTY_VALUES)


@njit(nogil=True)
def mult_vec(a_h, x):
    y = np.zeros(a_h.nrows, dtype=np.float64)

    if a_h.H:
        _x = ffi.from_buffer(x)
        _y = ffi.from_buffer(y)

        lk_mkl_spmv(1.0, a_h.H, _x, 0.0, _y)

    return y



import numpy as np
import scipy.sparse as sps

from csr import CSR

from pytest import mark


@mark.benchmark(
    group='MultAB'
)
def test_mult_ab(kernel, benchmark):
    A = sps.random(100, 500, 0.1, format='csr')
    B = sps.random(500, 200, 0.2, format='csr')
    A = CSR.from_scipy(A)
    B = CSR.from_scipy(B)

    # make sure it's compiled
    A.multiply(B)

    def op():
        A.multiply(B)

    benchmark(op)


@mark.parametrize('density', np.linspace(0, 1, 11))
@mark.benchmark(
    group='MultAB-Density'
)
def test_mult_ab_by_density(kernel, benchmark, density):
    A = sps.random(100, 100, density, format='csr')
    B = sps.random(100, 100, density, format='csr')
    A = CSR.from_scipy(A)
    B = CSR.from_scipy(B)

    # make sure it's compiled
    A.multiply(B)

    def op():
        A.multiply(B)

    benchmark(op)


@mark.parametrize('size', [5, 10, 15, 25, 50, 100, 200, 250, 500, 750, 1000])
@mark.benchmark(
    group='MultAB-Size',
    max_time=10
)
def test_mult_ab_by_size(kernel, benchmark, size):
    A = sps.random(size, size, 0.1, format='csr')
    B = sps.random(size, size, 0.1, format='csr')
    A = CSR.from_scipy(A)
    B = CSR.from_scipy(B)

    # make sure it's compiled
    A.multiply(B)

    def op():
        A.multiply(B)

    benchmark(op)


@mark.benchmark(
    group='MultABt'
)
def test_mult_abt(kernel, benchmark):
    A = sps.random(100, 500, 0.1, format='csr')
    B = sps.random(200, 500, 0.2, format='csr')
    A = CSR.from_scipy(A)
    B = CSR.from_scipy(B)

    # make sure it's compiled
    A.multiply(B, transpose=True)

    def op():
        A.multiply(B, transpose=True)

    benchmark(op)


@mark.parametrize('density', np.linspace(0, 1, 11))
@mark.benchmark(
    group='MultABt-Density',
)
def test_mult_abt_by_density(kernel, benchmark, density):
    A = sps.random(100, 100, density, format='csr')
    B = sps.random(100, 100, density, format='csr')
    A = CSR.from_scipy(A)
    B = CSR.from_scipy(B)

    # make sure it's compiled
    A.multiply(B, transpose=True)

    def op():
        A.multiply(B, transpose=True)

    benchmark(op)



import numpy as np
import scipy.sparse as sps

from csr import CSR
from csr.test_utils import csrs, matrices, sparse_matrices

from pytest import raises
from hypothesis import given, assume, settings, HealthCheck
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph


def test_csr_rowinds():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)
    csr = CSR.from_coo(rows, cols, vals)

    ris = csr.rowinds()
    assert all(ris == rows)


def test_csr_str():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)

    assert '4x3' in str(csr)


def test_csr_row_extent_fixed():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_) + 1
    csr = CSR.from_coo(rows, cols, vals)

    assert csr.row_extent(0) == (0, 2)
    assert csr.row_extent(1) == (2, 3)
    assert csr.row_extent(2) == (3, 3)
    assert csr.row_extent(3) == (3, 4)


@given(sparse_matrices())
def test_csr_row_extent(smat):
    csr = CSR.from_scipy(smat)

    for i in range(csr.nrows):
        sp, ep = csr.row_extent(i)
        assert sp == csr.rowptrs[i]
        assert ep == csr.rowptrs[i+1]


def test_csr_row_fixed():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_) + 1

    csr = CSR.from_coo(rows, cols, vals)
    assert all(csr.row(0) == np.array([0, 1, 2], dtype=np.float_))
    assert all(csr.row(1) == np.array([3, 0, 0], dtype=np.float_))
    assert all(csr.row(2) == np.array([0, 0, 0], dtype=np.float_))
    assert all(csr.row(3) == np.array([0, 4, 0], dtype=np.float_))


@settings(deadline=500)
@given(sparse_matrices())
def test_csr_row(smat):
    csr = CSR.from_scipy(smat)

    for i in range(csr.nrows):
        other = smat[i, :].toarray().ravel()
        row = csr.row(i)

        assert row.size == other.size

        assert all(row == other)


def test_csr_sparse_row():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)
    assert all(csr.row_cs(0) == np.array([1, 2], dtype=np.int32))
    assert all(csr.row_cs(1) == np.array([0], dtype=np.int32))
    assert all(csr.row_cs(2) == np.array([], dtype=np.int32))
    assert all(csr.row_cs(3) == np.array([1], dtype=np.int32))

    assert all(csr.row_vs(0) == np.array([0, 1], dtype=np.float_))
    assert all(csr.row_vs(1) == np.array([2], dtype=np.float_))
    assert all(csr.row_vs(2) == np.array([], dtype=np.float_))
    assert all(csr.row_vs(3) == np.array([3], dtype=np.float_))


@given(csrs())
def test_drop_values(csr):
    csr.drop_values()
    assert csr.values is None


@given(csrs(), st.floats(allow_infinity=False, allow_nan=False))
def test_fill_values(csr, x):
    csr.fill_values(x)
    assert all(csr.values == x)


def test_csr_set_values():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)

    v2 = np.random.randn(4)
    csr.values = v2

    assert all(csr.values == v2)


def test_csr_set_values_oversize():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)

    v2 = np.random.randn(6)
    csr.values = v2

    assert csr.values is not None
    assert csr.R.has_values
    assert all(csr.values == v2[:4])


def test_csr_set_values_undersize():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)

    v2 = np.random.randn(3)

    with raises(ValueError):
        csr.values = v2

    assert all(csr.values == vals)


def test_csr_set_values_none():
    rows = np.array([0, 0, 1, 3], dtype=np.int32)
    cols = np.array([1, 2, 0, 1], dtype=np.int32)
    vals = np.arange(4, dtype=np.float_)

    csr = CSR.from_coo(rows, cols, vals)
    csr.values = None

    assert csr.values is None
    assert all(csr.row(0) == [0, 1, 1])
    assert all(csr.row(1) == [1, 0, 0])
    assert all(csr.row(3) == [0, 1, 0])


@given(matrices())
def test_csr_row_nnzs(mat):
    nrows, ncols = mat.shape

    # sparsify the matrix
    mat[mat <= 0] = 0
    smat = sps.csr_matrix(mat)
    # make sure it's sparse
    assume(smat.nnz == np.sum(mat > 0))
    csr = CSR.from_scipy(smat)

    nnzs = csr.row_nnzs()
    assert nnzs.sum() == csr.nnz
    for i in range(nrows):
        row = mat[i, :]
        assert nnzs[i] == np.sum(row > 0)


@given(sparse_matrices())
def test_copy(mat):
    "Test copying a CSR"
    csr = CSR.from_scipy(mat)
    c2 = csr.copy()

    assert c2.nrows == csr.nrows
    assert c2.ncols == csr.ncols
    assert c2.nnz == csr.nnz
    assert c2.rowptrs is not csr.rowptrs
    assert all(c2.rowptrs == csr.rowptrs)
    assert c2.colinds is not csr.colinds
    assert all(c2.colinds == csr.colinds)
    assert c2.values is not csr.values
    assert all(c2.values == csr.values)


@given(sparse_matrices())
def test_copy_share(mat):
    "Test copying a CSR and sharing structure"
    csr = CSR.from_scipy(mat)
    c2 = csr.copy(copy_structure=False)

    assert c2.nrows == csr.nrows
    assert c2.ncols == csr.ncols
    assert c2.nnz == csr.nnz
    assert c2.rowptrs is csr.rowptrs
    assert c2.colinds is csr.colinds
    assert c2.values is not csr.values
    assert all(c2.values == csr.values)


@given(sparse_matrices())
def test_copy_structure_only(mat):
    "Test copying only the structure of a CSV."
    csr = CSR.from_scipy(mat)
    c2 = csr.copy(False)

    assert c2.nrows == csr.nrows
    assert c2.ncols == csr.ncols
    assert c2.nnz == csr.nnz
    assert c2.rowptrs is not csr.rowptrs
    assert all(c2.rowptrs == csr.rowptrs)
    assert c2.colinds is not csr.colinds
    assert all(c2.colinds == csr.colinds)
    assert c2.values is None


@given(csrs(values=False), st.booleans())
def test_copy_csrnv(csr, inc):
    "Test copying a CSR with no values."
    c2 = csr.copy(inc)

    assert c2.nrows == csr.nrows
    assert c2.ncols == csr.ncols
    assert c2.nnz == csr.nnz
    assert c2.rowptrs is not csr.rowptrs
    assert all(c2.rowptrs == csr.rowptrs)
    assert c2.colinds is not csr.colinds
    assert all(c2.colinds == csr.colinds)
    assert c2.values is None



import numpy as np
import scipy.sparse as sps

from csr import CSR

from pytest import mark


@mark.benchmark(
    group='MultAx'
)
def test_mult_vec(kernel, benchmark):
    A = sps.random(100, 100, 0.1, format='csr')
    A = CSR.from_scipy(A)
    x = np.random.randn(100)

    # make sure it's compiled
    y = A.mult_vec(x)
    assert len(y) == A.nrows

    def op():
        A.mult_vec(x)

    benchmark(op)



"""
Backend implementations of Numba operations.
"""

import logging
import numpy as np
from numba import njit, prange

from .layout import _CSR, EMPTY_VALUES

_log = logging.getLogger(__name__)


@njit
def _swap(a, i, j):
    a[i], a[j] = a[j], a[i]


@njit
def make_empty(nrows, ncols):
    rowptrs = np.zeros(nrows + 1, dtype=np.intc)
    colinds = np.zeros(0, dtype=np.intc)
    values = np.zeros(0)
    return _CSR(nrows, ncols, 0, rowptrs, colinds, values)


@njit
def make_unintialized(nrows, ncols, sizes):
    nnz = np.sum(sizes)
    rowptrs = np.zeros(nrows + 1, dtype=np.intc)
    for i in range(nrows):
        rowptrs[i+1] = rowptrs[i] + sizes[i]
    colinds = np.full(nnz, -1, dtype=np.intc)
    values = np.full(nnz, np.nan)
    return _CSR(nrows, ncols, nnz, rowptrs, colinds, values)


@njit
def row_extent(csr, row):
    "Get the extent of a row in the matrix storage."
    sp = csr.rowptrs[row]
    ep = csr.rowptrs[row+1]
    return sp, ep


@njit
def row(csr, row):
    "Get a row as a dense vector."
    v = np.zeros(csr.ncols)
    if csr.nnz == 0:
        return v

    sp, ep = row_extent(csr, row)
    cols = csr.colinds[sp:ep]
    if csr.has_values > 0:
        v[cols] = csr.values[sp:ep]
    else:
        v[cols] = 1

    return v


@njit
def row_cs(csr, row):
    "Get the column indices for a row."
    sp = csr.rowptrs[row]
    ep = csr.rowptrs[row + 1]

    return csr.colinds[sp:ep]


@njit
def row_vs(csr, row):
    "Get the nonzero values for a row."
    sp = csr.rowptrs[row]
    ep = csr.rowptrs[row + 1]

    if csr.has_values:
        return csr.values[sp:ep]
    else:
        return np.full(ep - sp, 1.0)


@njit
def rowinds(csr):
    "Get the row indices for the nonzero values in a matrix."
    ris = np.zeros(csr.nnz, np.intc)
    for i in range(csr.nrows):
        sp, ep = row_extent(csr, i)
        ris[sp:ep] = i
    return ris


@njit
def subset_rows(csr, begin, end):
    "Take a subset of the rows of a CSR."
    st = csr.rowptrs[begin]
    ed = csr.rowptrs[end]
    rps = csr.rowptrs[begin:(end+1)] - st

    cis = csr.colinds[st:ed]
    if csr.has_values:
        vs = csr.values[st:ed]
    else:
        vs = None
    return _CSR(end - begin, csr.ncols, ed - st, rps, cis, vs)


@njit(nogil=True)
def center_rows(csr):
    "Mean-center the nonzero values of each row of a CSR."
    means = np.zeros(csr.nrows)
    for i in range(csr.nrows):
        sp, ep = row_extent(csr, i)
        if sp == ep:
            continue  # empty row
        vs = row_vs(csr, i)
        m = np.mean(vs)
        means[i] = m
        csr.values[sp:ep] -= m

    return means


@njit(nogil=True)
def unit_rows(csr):
    "Normalize the rows of a CSR to unit vectors."
    norms = np.zeros(csr.nrows)
    for i in range(csr.nrows):
        sp, ep = row_extent(csr, i)
        if sp == ep:
            continue  # empty row
        vs = row_vs(csr, i)
        m = np.linalg.norm(vs)
        norms[i] = m
        csr.values[sp:ep] /= m

    return norms


@njit(nogil=True)
def transpose(csr, include_values):
    "Transpose a CSR."
    brp = np.zeros(csr.ncols + 1, csr.rowptrs.dtype)
    bci = np.zeros(csr.nnz, np.int32)
    if include_values and csr.has_values:
        bvs = np.zeros(csr.nnz, np.float64)
    else:
        bvs = np.zeros(0)

    # count elements
    for i in range(csr.nrows):
        ars, are = row_extent(csr, i)
        for jj in range(ars, are):
            j = csr.colinds[jj]
            brp[j+1] += 1

    # convert to pointers
    for j in range(csr.ncols):
        brp[j+1] = brp[j] + brp[j+1]

    # construct results
    for i in range(csr.nrows):
        ars, are = row_extent(csr, i)
        for jj in range(ars, are):
            j = csr.colinds[jj]
            bci[brp[j]] = i
            if include_values and csr.has_values:
                bvs[brp[j]] = csr.values[jj]
            brp[j] += 1

    # restore pointers
    for i in range(csr.ncols - 1, 0, -1):
        brp[i] = brp[i-1]
    brp[0] = 0

    if not include_values or not csr.has_values:
        bvs = None

    return _CSR(csr.ncols, csr.nrows, csr.nnz, brp, bci, bvs)


@njit(nogil=True)
def sort_rows(csr):
    "Sort the rows of a CSR by increasing column index"
    for i in range(csr.nrows):
        sp, ep = row_extent(csr, i)
        # bubble-sort so it's super-fast on sorted arrays
        swapped = True
        while swapped:
            swapped = False
            for j in range(sp, ep - 1):
                if csr.colinds[j] > csr.colinds[j+1]:
                    _swap(csr.colinds, j, j+1)
                    if csr.has_values:
                        _swap(csr.values, j, j+1)
                    swapped = True


@njit(nogil=True)
def _csr_align(rowinds, nrows, rowptrs, align):
    rcts = np.zeros(nrows, dtype=rowptrs.dtype)
    for r in rowinds:
        rcts[r] += 1

    rowptrs[1:] = np.cumsum(rcts)
    rpos = rowptrs[:-1].copy()

    for i in range(len(rowinds)):
        row = rowinds[i]
        pos = rpos[row]
        align[pos] = i
        rpos[row] += 1



"""
Type for CSR storage.
"""

import numpy as np
from numba.experimental import jitclass
import numba.types as nt


@jitclass([
    ('nrows', nt.intc),
    ('ncols', nt.intc),
    ('nnz', nt.intc),
    ('has_values', nt.boolean),
    ('rowptrs', nt.intc[::1]),
    ('colinds', nt.intc[::1]),
    ('values', nt.float64[::1])
])
class _CSR:
    """
    Internal storage for :py:class:`csr.CSR`. If you work with CSRs from Numba,
    you will use this type instead of the CSR class itself, with functions
    from :py:mod:`csr.native_ops`.

    This has the same key attributes as :py:class:`csr.CSR`, except that it always
    has a ``values`` array; if only structure is stored, this array has length 0.

    Attributes:
        nrows(int): the number of rows
        ncols(int): the number of columns
        nnz(int): the number of nonzero entries
        rowptrs(numpy.ndarray): starting position of each row (length ``nrows + 1``)
        colinds(numpy.ndarray): column indices (length ``nnz``)
        has_values(bool):
            Whether or not this matrix has values, or only stores the structure.
        values(numpy.ndarray):
            matrix cell values (length ``nnz`` or 0). If only the matrix structure
            is stored, this has length 0.
    """
    def __init__(self, nrows, ncols, nnz, rowptrs, colinds, values):
        self.nrows = nrows
        self.ncols = ncols
        self.nnz = nnz
        self.rowptrs = rowptrs
        self.colinds = colinds
        if values is None:
            self.values = EMPTY_VALUES
            self.has_values = False
        else:
            self.values = values
            self.has_values = True


EMPTY_VALUES = np.zeros(0)



"""
Tests for creating kernel handles.
"""

from csr import CSR
from csr.test_utils import csrs, sparse_matrices

from hypothesis import given, settings


@settings(deadline=5000)
@given(sparse_matrices())
def test_make_handle(kernel, mat):
    csr = CSR.from_scipy(mat)
    h = kernel.to_handle(csr.R)
    try:
        assert h is not None
        c2 = kernel.from_handle(h)
        assert c2.nrows == csr.nrows
        assert c2.ncols == csr.ncols
        assert c2.nnz == csr.nnz
    finally:
        kernel.release_handle(h)



import logging

from csr import CSR
from csr.test_utils import mm_pairs

from pytest import approx
from hypothesis import given, settings

_log = logging.getLogger(__name__)


@settings(deadline=None)
@given(mm_pairs())
def test_multiply(kernel, pair):
    A, B = pair
    csra = CSR.from_scipy(A)
    csrb = CSR.from_scipy(B)

    prod = csra.multiply(csrb)
    assert isinstance(prod, CSR)
    _log.info('got %r', prod)
    _log.info('inner: %s', prod.R)
    assert prod.nrows == csra.nrows
    assert prod.ncols == csrb.ncols

    AB = A @ B
    abnr, abnc = AB.shape
    assert prod.nrows == abnr
    assert prod.ncols == abnc

    assert prod.nnz == AB.nnz
    if prod.nnz > 0:
        assert prod.values is not None
    nrows = csra.nrows

    for i in range(nrows):
        r_scipy = AB.getrow(i).toarray().ravel()
        r_ours = prod.row(i)

        assert len(r_ours) == len(r_scipy)
        assert r_ours == approx(r_scipy)


@settings(deadline=None)
@given(mm_pairs())
def test_multiply_transpose(kernel, pair):
    A, B = pair
    csra = CSR.from_scipy(A)
    csrb = CSR.from_scipy(B.T)

    prod = csra.multiply(csrb, transpose=True)
    assert isinstance(prod, CSR)
    _log.info('got %r', prod)
    _log.info('inner: %s', prod.R)
    assert prod.nrows == csra.nrows
    assert prod.ncols == csrb.nrows

    AB = A @ B
    abnr, abnc = AB.shape
    assert prod.nrows == abnr
    assert prod.ncols == abnc

    assert prod.nnz == AB.nnz
    if prod.nnz > 0:
        assert prod.values is not None
    nrows = csra.nrows

    for i in range(nrows):
        r_scipy = AB.getrow(i).toarray().ravel()
        r_ours = prod.row(i)

        assert len(r_ours) == len(r_scipy)
        assert r_ours == approx(r_scipy)



from types import BuiltinFunctionType
import numba.types as nt
from numba.core.typing.cffi_utils import register_module, register_type
from numba import njit

from . import _mkl_ops

register_type(_mkl_ops.ffi.typeof('intptr_t'), nt.intp)
register_module(_mkl_ops)

ffi = _mkl_ops.ffi

# export all the LK names
__all__ = ['ffi']
for name in dir(_mkl_ops.lib):
    f = getattr(_mkl_ops.lib, name)
    if isinstance(f, BuiltinFunctionType):
        globals()[name] = f
        __all__.append(name)
